"""A simple example to show message passing between jobs."""
from jobflow import Flow, job, run_locally


@job
def encode_message(message):
    """Encode a message using base64."""
    from base64 import b64encode

    return b64encode(message.encode()).decode()


@job
def decode_message(message):
    """Decode a message from base64."""
    from base64 import b64decode

    return b64decode(message.encode()).decode()


# Create two jobs, the first to encode a message and the second to decode it.
encode = encode_message("Lo, a shadow of horror is risen")
decode = decode_message(encode.output)

# Create a flow containing the jobs. The order of the jobs doesn't matter and will be
# determined by the connectivity of the jobs.
flow = Flow([encode, decode])

# draw the flow graph
flow.draw_graph().show()

# run the flow, "output" contains the output of all jobs
output = run_locally(flow)
print(output)



import pytest


@pytest.fixture(scope="session")
def test_data():
    from pathlib import Path

    module_dir = Path(__file__).resolve().parent
    test_dir = module_dir / "test_data"
    return test_dir.resolve()


@pytest.fixture(scope="session")
def database():
    return "jobflow_test"


@pytest.fixture(scope="session")
def mongo_jobstore(database):
    from maggma.stores import MongoStore

    from jobflow import JobStore

    store = JobStore(MongoStore(database, "outputs"))
    store.connect()
    return store


@pytest.fixture(scope="function")
def memory_jobstore():
    from maggma.stores import MemoryStore

    from jobflow import JobStore

    store = JobStore(MemoryStore())
    store.connect()

    return store


@pytest.fixture(scope="function")
def memory_data_jobstore():
    from maggma.stores import MemoryStore

    from jobflow import JobStore

    store = JobStore(MemoryStore(), additional_stores={"data": MemoryStore()})
    store.connect()

    return store


@pytest.fixture
def clean_dir():
    import os
    import shutil
    import tempfile

    old_cwd = os.getcwd()
    newpath = tempfile.mkdtemp()
    os.chdir(newpath)

    yield

    os.chdir(old_cwd)
    shutil.rmtree(newpath)


@pytest.fixture(scope="session")
def debug_mode():
    return False


@pytest.fixture(scope="session")
def lpad(database, debug_mode):
    from fireworks import LaunchPad

    lpad = LaunchPad(name=database)
    lpad.reset("", require_password=False)
    yield lpad

    if not debug_mode:
        lpad.reset("", require_password=False)
        for coll in lpad.db.list_collection_names():
            lpad.db[coll].drop()


@pytest.fixture
def no_pydot(monkeypatch):
    import builtins

    import_orig = builtins.__import__

    def mocked_import(name, *args, **kwargs):
        if name == "pydot":
            raise ImportError()
        return import_orig(name, *args, **kwargs)

    monkeypatch.setattr(builtins, "__import__", mocked_import)


@pytest.fixture
def no_matplotlib(monkeypatch):
    import builtins

    import_orig = builtins.__import__

    def mocked_import(name, *args, **kwargs):
        if name == "matplotlib":
            raise ImportError()
        return import_orig(name, *args, **kwargs)

    monkeypatch.setattr(builtins, "__import__", mocked_import)



import pytest


def add(a, b):
    return a + b


def div(a, b=2):
    return a / b


def get_test_job():
    from jobflow import Job

    return Job(add, function_args=(1, 2))


def get_test_flow():
    from jobflow import Flow, Job

    add_job = Job(add, function_args=(1, 2))
    div_job = Job(div, function_args=(add_job.output,), function_kwargs={"b": 3})
    return Flow([add_job, div_job])


def get_maker_flow(return_makers=False):
    from dataclasses import dataclass

    from jobflow import Flow, Maker, job

    @dataclass
    class AddMaker(Maker):
        name: str = "add"
        b: int = 2

        @job
        def make(self, a):
            return a + self.b

    @dataclass
    class DivMaker(Maker):
        name: str = "div"
        b: int = 5

        @job
        def make(self, a):
            return a + self.b

    add_maker = AddMaker(b=3)
    div_maker = DivMaker(b=4)
    add_job = add_maker.make(2)
    div_job = div_maker.make(add_job.output)
    flow = Flow([add_job, div_job])

    if return_makers:
        return flow, (AddMaker, DivMaker)
    else:
        return flow


def test_flow_of_jobs_init():
    from jobflow.core.flow import Flow, JobOrder

    # test single job
    add_job = get_test_job()
    flow = Flow([add_job], name="add")
    assert flow.name == "add"
    assert flow.host is None
    assert flow.output is None
    assert flow.job_uuids == (add_job.uuid,)

    # test single job no list
    add_job = get_test_job()
    flow = Flow(add_job, name="add")
    assert flow.name == "add"
    assert flow.host is None
    assert flow.output is None
    assert flow.job_uuids == (add_job.uuid,)

    # # test multiple job
    add_job1 = get_test_job()
    add_job2 = get_test_job()
    flow = Flow([add_job1, add_job2])
    assert flow.host is None
    assert flow.output is None
    assert flow.job_uuids == (add_job1.uuid, add_job2.uuid)

    # # test multiple job, linear order
    add_job1 = get_test_job()
    add_job2 = get_test_job()
    flow = Flow([add_job1, add_job2], order=JobOrder.LINEAR)
    assert flow.host is None
    assert flow.output is None
    assert flow.job_uuids == (add_job1.uuid, add_job2.uuid)

    # test single job and output
    add_job = get_test_job()
    flow = Flow([add_job], output=add_job.output)
    assert flow.output == add_job.output

    # # test multi job and list multi outputs
    add_job1 = get_test_job()
    add_job2 = get_test_job()
    flow = Flow([add_job1, add_job2], output=[add_job1.output, add_job2.output])
    assert flow.output[1] == add_job2.output

    # # test all jobs included needed to generate outputs
    add_job = get_test_job()
    with pytest.raises(ValueError):
        Flow([], output=add_job.output)

    # test job given rather than outputs
    add_job = get_test_job()
    with pytest.warns(UserWarning):
        Flow([add_job], output=add_job)

    # test complex object containing job given rather than outputs
    add_job = get_test_job()
    with pytest.warns(UserWarning):
        Flow([add_job], output={1: [[{"a": add_job}]]})

    # test job already belongs to another flow
    add_job = get_test_job()
    Flow([add_job])
    with pytest.raises(ValueError):
        Flow([add_job])

    # test that two of the same job cannot be used in the same flow
    add_job = get_test_job()
    with pytest.raises(ValueError):
        Flow([add_job, add_job])


def test_flow_of_flows_init():
    from jobflow.core.flow import Flow

    # test single flow
    add_job = get_test_job()
    subflow = Flow([add_job])
    flow = Flow([subflow], name="add")
    assert flow.name == "add"
    assert flow.host is None
    assert flow.output is None
    assert flow.job_uuids == (add_job.uuid,)
    assert flow.jobs[0].host == flow.uuid

    # test single flow no list
    add_job = get_test_job()
    subflow = Flow(add_job)
    flow = Flow(subflow, name="add")
    assert flow.name == "add"
    assert flow.host is None
    assert flow.output is None
    assert flow.job_uuids == (add_job.uuid,)
    assert flow.jobs[0].host == flow.uuid

    # test multiple flows
    add_job1 = get_test_job()
    add_job2 = get_test_job()
    subflow1 = Flow([add_job1])
    subflow2 = Flow([add_job2])
    flow = Flow([subflow1, subflow2])
    assert flow.host is None
    assert flow.output is None
    assert flow.job_uuids == (add_job1.uuid, add_job2.uuid)
    assert flow.jobs[0].host == flow.uuid
    assert flow.jobs[1].host == flow.uuid

    # test single job and outputs
    add_job = get_test_job()
    subflow = Flow([add_job], output=add_job.output)
    flow = Flow([subflow], output=subflow.output)
    assert flow.output == add_job.output

    # test multi job and list multi outputs
    add_job1 = get_test_job()
    add_job2 = get_test_job()
    subflow1 = Flow([add_job1], output=add_job1.output)
    subflow2 = Flow([add_job2], output=add_job2.output)
    flow = Flow([subflow1, subflow2], output=[subflow1.output, subflow2.output])
    assert flow.output[0] == add_job1.output
    assert flow.output[1] == add_job2.output

    # test all jobflow included needed to generate outputs
    add_job = get_test_job()
    subflow = Flow([add_job], output=add_job.output)
    with pytest.raises(ValueError):
        Flow([], output=subflow.output)

    # test flow given rather than outputs
    add_job = get_test_job()
    subflow = Flow([add_job], output=add_job.output)
    with pytest.warns(UserWarning):
        Flow([subflow], output=subflow)

    # test complex object containing job given rather than outputs
    add_job = get_test_job()
    subflow = Flow([add_job], output=add_job.output)
    with pytest.warns(UserWarning):
        Flow([subflow], output={1: [[{"a": subflow}]]})

    # test flow already belongs to another flow
    add_job = get_test_job()
    subflow = Flow([add_job], output=add_job.output)
    Flow([subflow])
    with pytest.raises(ValueError):
        Flow([subflow])

    # test that two of the same flow cannot be used in the same flow
    add_job = get_test_job()
    subflow = Flow([add_job], output=add_job.output)
    with pytest.raises(ValueError):
        Flow([subflow, subflow])


def test_flow_job_mixed():
    from jobflow.core.flow import Flow

    # test job and flows
    add_job = get_test_job()
    add_job2 = get_test_job()
    subflow = Flow([add_job2])
    flow = Flow([add_job, subflow])
    assert flow.host is None
    assert flow.output is None
    assert flow.job_uuids == (add_job.uuid, add_job2.uuid)
    assert flow.jobs[0].host == flow.uuid
    assert flow.jobs[1].host == flow.uuid

    # test with list multi outputs
    add_job = get_test_job()
    add_job2 = get_test_job()
    subflow = Flow([add_job2], output=add_job2.output)
    flow = Flow([add_job, subflow], output=[add_job.output, subflow.output])
    assert flow.output[0] == add_job.output
    assert flow.output[1] == add_job2.output

    # test all jobs/flows included needed to generate outputs
    add_job = get_test_job()
    add_job2 = get_test_job()
    subflow = Flow([add_job2], output=add_job2.output)
    with pytest.raises(ValueError):
        Flow([add_job], output=[add_job.output, subflow.output])


def test_graph():
    from jobflow import Flow, JobOrder

    # test unconnected graph
    add_job1 = get_test_job()
    add_job2 = get_test_job()
    flow = Flow([add_job1, add_job2])
    graph = flow.graph
    assert len(graph.edges) == 0
    assert len(graph.nodes) == 2

    # test unconnected graph, linear order
    add_job1 = get_test_job()
    add_job2 = get_test_job()
    flow = Flow([add_job1, add_job2], order=JobOrder.LINEAR)
    graph = flow.graph
    assert len(graph.edges) == 1
    assert len(graph.nodes) == 2

    # test connected graph, wrong order
    add_job1 = get_test_job()
    add_job2 = get_test_job()
    add_job1.function_args = (2, add_job2.output)
    flow = Flow([add_job1, add_job2])
    graph = flow.graph
    assert len(graph.edges) == 1
    assert len(graph.nodes) == 2

    # test connected graph, linear order
    add_job1 = get_test_job()
    add_job2 = get_test_job()
    add_job1.function_args = (2, add_job2.output)
    flow = Flow([add_job1, add_job2], order=JobOrder.LINEAR)
    graph = flow.graph
    assert len(graph.edges) == 2
    assert len(graph.nodes) == 2

    # test unconnected graph
    add_job1 = get_test_job()
    add_job2 = get_test_job()
    add_job3 = get_test_job()
    add_job4 = get_test_job()
    subflow1 = Flow([add_job1, add_job2])
    subflow2 = Flow([add_job3, add_job4])
    flow = Flow([subflow1, subflow2])
    graph = flow.graph
    assert len(graph.edges) == 0
    assert len(graph.nodes) == 4

    # test unconnected graph, linear order
    add_job1 = get_test_job()
    add_job2 = get_test_job()
    add_job3 = get_test_job()
    add_job4 = get_test_job()
    subflow1 = Flow([add_job1, add_job2])
    subflow2 = Flow([add_job3, add_job4])
    flow = Flow([subflow1, subflow2], order=JobOrder.LINEAR)
    graph = flow.graph
    assert len(graph.edges) == 4
    assert len(graph.nodes) == 4

    # test connected graph, wrong order
    add_job1 = get_test_job()
    add_job2 = get_test_job()
    add_job3 = get_test_job()
    add_job4 = get_test_job()
    subflow1 = Flow([add_job1, add_job2])
    subflow2 = Flow([add_job3, add_job4])
    add_job1.function_args = (2, add_job3.output)
    flow = Flow([subflow1, subflow2])
    graph = flow.graph
    assert len(graph.edges) == 1
    assert len(graph.nodes) == 4

    # test connected graph, linear order
    add_job1 = get_test_job()
    add_job2 = get_test_job()
    add_job3 = get_test_job()
    add_job4 = get_test_job()
    subflow1 = Flow([add_job1, add_job2])
    subflow2 = Flow([add_job3, add_job4])
    add_job1.function_args = (2, add_job3.output)
    flow = Flow([subflow1, subflow2], order=JobOrder.LINEAR)
    graph = flow.graph
    assert len(graph.edges) == 5
    assert len(graph.nodes) == 4


def test_draw_graph():
    from jobflow import Flow, JobOrder

    # test unconnected graph
    add_job1 = get_test_job()
    add_job2 = get_test_job()
    flow = Flow([add_job1, add_job2])
    assert flow.draw_graph()

    # test unconnected graph, linear order
    add_job1 = get_test_job()
    add_job2 = get_test_job()
    flow = Flow([add_job1, add_job2], order=JobOrder.LINEAR)
    assert flow.draw_graph()

    # test connected graph, wrong order
    add_job1 = get_test_job()
    add_job2 = get_test_job()
    add_job1.function_args = (2, add_job2.output)
    flow = Flow([add_job1, add_job2])
    assert flow.draw_graph()

    # test connected graph, linear order
    add_job1 = get_test_job()
    add_job2 = get_test_job()
    add_job1.function_args = (2, add_job2.output)
    flow = Flow([add_job1, add_job2], order=JobOrder.LINEAR)
    assert flow.draw_graph()


@pytest.mark.usefixtures("no_pydot")
def test_draw_graph_nopydot():
    from jobflow import Flow, JobOrder

    # test unconnected graph
    add_job1 = get_test_job()
    add_job2 = get_test_job()
    flow = Flow([add_job1, add_job2])
    assert flow.draw_graph()

    # test unconnected graph, linear order
    add_job1 = get_test_job()
    add_job2 = get_test_job()
    flow = Flow([add_job1, add_job2], order=JobOrder.LINEAR)
    assert flow.draw_graph()

    # test connected graph, wrong order
    add_job1 = get_test_job()
    add_job2 = get_test_job()
    add_job1.function_args = (2, add_job2.output)
    flow = Flow([add_job1, add_job2])
    assert flow.draw_graph()

    # test connected graph, linear order
    add_job1 = get_test_job()
    add_job2 = get_test_job()
    add_job1.function_args = (2, add_job2.output)
    flow = Flow([add_job1, add_job2], order=JobOrder.LINEAR)
    assert flow.draw_graph()


def test_iterflow():
    from jobflow import Flow, JobOrder

    # test unconnected graph
    add_job1 = get_test_job()
    add_job2 = get_test_job()
    flow = Flow([add_job1, add_job2])
    iterated = list(flow.iterflow())
    assert len(iterated) == 2
    assert iterated[0][0] == add_job1
    assert len(iterated[0][1]) == 0
    assert iterated[1][0] == add_job2
    assert len(iterated[1][1]) == 0

    # test unconnected graph, linear order
    add_job1 = get_test_job()
    add_job2 = get_test_job()
    flow = Flow([add_job1, add_job2], order=JobOrder.LINEAR)
    iterated = list(flow.iterflow())
    assert len(iterated) == 2
    assert iterated[0][0] == add_job1
    assert len(iterated[0][1]) == 0
    assert iterated[1][0] == add_job2
    assert len(iterated[1][1]) == 1

    # test connected graph, wrong order
    add_job1 = get_test_job()
    add_job2 = get_test_job()
    add_job1.function_args = (2, add_job2.output)
    flow = Flow([add_job1, add_job2])
    iterated = list(flow.iterflow())
    assert len(iterated) == 2
    assert iterated[0][0] == add_job2
    assert len(iterated[0][1]) == 0
    assert iterated[1][0] == add_job1
    assert len(iterated[1][1]) == 1

    # test connected graph, linear order
    add_job1 = get_test_job()
    add_job2 = get_test_job()
    add_job1.function_args = (2, add_job2.output)
    flow = Flow([add_job1, add_job2], order=JobOrder.LINEAR)
    with pytest.raises(ValueError):
        list(flow.iterflow())


def test_dag_validation():
    from jobflow import Flow, Job

    # test cycle detection of jobs
    job1 = Job(add, function_args=(1, 2))
    job2 = Job(add, function_args=(job1.output, 2))
    job1.function_args = (job2.output, 2)
    flow = Flow(jobs=[job1, job2])
    with pytest.raises(ValueError):
        next(flow.iterflow())


def test_serialization():
    import json

    from monty.json import MontyDecoder, MontyEncoder

    from jobflow import Flow

    flow = Flow([])
    flow_host = Flow([flow])
    host_uuid = flow_host.uuid

    encoded_flow = json.loads(MontyEncoder().encode(flow_host))
    decoded_flow = MontyDecoder().process_decoded(encoded_flow)

    assert decoded_flow.jobs[0].host == host_uuid


def test_update_kwargs():
    # test no filter
    flow = get_test_flow()
    flow.update_kwargs({"b": 5})
    assert flow.jobs[0].function_kwargs["b"] == 5
    assert flow.jobs[1].function_kwargs["b"] == 5

    # test name filter
    flow = get_test_flow()
    flow.update_kwargs({"b": 5}, name_filter="div")
    assert "b" not in flow.jobs[0].function_kwargs
    assert flow.jobs[1].function_kwargs["b"] == 5

    # test function filter
    flow = get_test_flow()
    flow.update_kwargs({"b": 5}, function_filter=div)
    assert "b" not in flow.jobs[0].function_kwargs
    assert flow.jobs[1].function_kwargs["b"] == 5

    # test dict mod
    flow = get_test_flow()
    flow.update_kwargs({"_inc": {"b": 5}}, function_filter=div, dict_mod=True)
    assert "b" not in flow.jobs[0].function_kwargs
    assert flow.jobs[1].function_kwargs["b"] == 8


def test_update_maker_kwargs():

    # test no filter
    flow = get_maker_flow()
    flow.update_maker_kwargs({"b": 10})
    assert flow.jobs[0].maker.b == 10
    assert flow.jobs[1].maker.b == 10

    # test bad kwarg
    flow = get_maker_flow()
    with pytest.raises(TypeError):
        flow.update_maker_kwargs({"c": 10})

    # test name filter
    flow = get_maker_flow()
    flow.update_maker_kwargs({"b": 10}, name_filter="div")
    assert flow.jobs[0].maker.b == 3
    assert flow.jobs[1].maker.b == 10

    # test class filter
    flow, (_, DivMaker) = get_maker_flow(return_makers=True)
    flow.update_maker_kwargs({"b": 10}, class_filter=DivMaker)
    assert flow.jobs[0].maker.b == 3
    assert flow.jobs[1].maker.b == 10

    # test class filter with instance
    flow, (_, DivMaker) = get_maker_flow(return_makers=True)
    div_maker = DivMaker()
    flow.update_maker_kwargs({"b": 10}, class_filter=div_maker)
    assert flow.jobs[0].maker.b == 3
    assert flow.jobs[1].maker.b == 10

    # test dict mod
    flow = get_maker_flow()
    flow.update_maker_kwargs({"_inc": {"b": 10}}, name_filter="div", dict_mod=True)
    assert flow.jobs[0].maker.b == 3
    assert flow.jobs[1].maker.b == 14


def test_append_name():
    from jobflow import Flow

    # test append
    flow = get_test_flow()
    flow.append_name(" test")
    assert flow.name == "Flow test"
    assert flow.jobs[0].name == "add test"

    # test prepend
    flow = get_test_flow()
    flow.append_name("test ", prepend=True)
    assert flow.name == "test Flow"
    assert flow.jobs[0].name == "test add"

    # test empty Flow
    flow = Flow([], name="abc")
    flow.append_name(" test")
    assert flow.name == "abc test"


def test_get_flow():
    import jobflow
    from jobflow import Job
    from jobflow.core.flow import get_flow

    # test get_flow method with a single job
    job = get_test_job()
    assert isinstance(job, jobflow.Job)
    flow = get_flow(job)
    assert isinstance(flow, jobflow.Flow)
    assert flow.jobs[0] is job

    # test get_flow method with a list of jobs
    job1 = get_test_job()
    job2 = get_test_job()
    flow = get_flow([job1, job2])
    assert isinstance(flow, jobflow.Flow)
    assert len(flow.jobs) == 2
    assert flow.jobs[0] is job1
    assert flow.jobs[1] is job2

    # test get_flow method with a flow
    flw = get_test_flow()
    assert isinstance(flw, jobflow.Flow)
    flow = get_flow(flw)
    assert isinstance(flow, jobflow.Flow)
    assert flow is flw

    # test all jobs included for graph to work
    job1 = Job(add, function_args=(1, 2))
    job2 = Job(add, function_args=(job1.output.value, 2))
    with pytest.raises(ValueError):
        get_flow(job2)

    # test all jobs included for graph to work
    job1 = Job(add, function_args=(1, 2))
    job2 = Job(add, function_args=(job1.output.value, 2))
    get_flow([job1, job2])



import pytest


def add(a, b=5):
    return a + b


def bad_output():
    return {1, 2, 3}


def test_job_init():
    from jobflow.core.job import Job

    # test basic init
    test_job = Job(function=add, function_args=("I am a job",))
    assert test_job
    assert test_job.function == add
    assert test_job.name == "add"
    assert test_job.function_args == ("I am a job",)
    assert test_job.function_kwargs == {}
    assert test_job.uuid is not None
    assert test_job.output.uuid == test_job.uuid

    # test init no args
    test_job = Job(function=add, function_args=tuple())
    assert test_job
    assert test_job.function == add
    assert test_job.name == "add"
    assert test_job.function_args == tuple()
    assert test_job.function_kwargs == {}
    assert test_job.uuid is not None
    assert test_job.output.uuid == test_job.uuid

    # test job as another job as input
    with pytest.warns(UserWarning):
        Job(function=add, function_args=(test_job,))

    # test init with kwargs
    test_job = Job(function=add, function_args=(1,), function_kwargs={"b": 2})
    assert test_job
    assert test_job.function == add
    assert test_job.name == "add"
    assert test_job.function_args == (1,)
    assert test_job.function_kwargs == {"b": 2}
    assert test_job.uuid is not None
    assert test_job.uuid == test_job.output.value.uuid

    # test init with stores
    test_job = Job(
        function=add,
        function_args=(1,),
        function_kwargs={"b": 2},
        data="output",
        graphs="graph",
    )
    assert test_job
    assert test_job.uuid == test_job.output.value.uuid
    assert test_job._kwargs == {"data": "output", "graphs": "graph"}

    # check giving True for multiple stores fails
    with pytest.raises(ValueError):
        Job(function=add, function_args=(1,), data=True, graphs=True)

    # test changing job name (test needed due to setattr override
    test_job = Job(function=add, function_args=("I am a job",))
    test_job.name = "abc"
    assert test_job.name == "abc"

    test_job.name += "xyz"
    assert test_job.name == "abcxyz"


def test_job_run(capsys, memory_jobstore, memory_data_jobstore):
    from jobflow.core.job import Job, Response

    # test basic run
    test_job = Job(print, function_args=("I am a job",))
    response = test_job.run(memory_jobstore)
    assert capsys.readouterr().out == "I am a job\n"
    assert isinstance(response, Response)

    # test run with outputs
    test_job = Job(add, function_args=(1,), function_kwargs={"b": 2})
    response = test_job.run(memory_jobstore)
    assert isinstance(response, Response)
    assert response.output == 3

    # test run with input references
    test_job = Job(add, function_args=(test_job.output,))
    response = test_job.run(memory_jobstore)
    assert isinstance(response, Response)
    assert response.output == 8

    def add_response(a, b):
        return Response(output=a + b, stop_children=True)

    test_job = Job(add_response, function_args=(1, 2))
    response = test_job.run(memory_jobstore)
    assert isinstance(response, Response)
    assert response.output == 3
    assert response.stop_children

    # test run with outputs and data store
    test_job = Job(add, function_args=(1,), function_kwargs={"b": 2}, data=True)
    response = test_job.run(memory_data_jobstore)
    assert isinstance(response, Response)
    assert response.output == 3

    # check output was not stored in the docs store
    result = memory_data_jobstore.query_one({"uuid": test_job.uuid})
    assert isinstance(result["output"], dict)
    assert "blob_uuid" in result["output"]

    # check the output can be resolved
    result = memory_data_jobstore.query_one({"uuid": test_job.uuid}, load=True)
    assert result["output"] == 3

    # test non MSONable output
    test_job = Job(bad_output)
    with pytest.raises(RuntimeError):
        test_job.run(memory_jobstore)


def test_replace_response(memory_jobstore):
    from jobflow import Flow, Job, Response

    def replace_job():
        job = Job(add, function_args=(1,))
        return Response(replace=job)

    def replace_list_job():
        job1 = Job(add, function_args=(1,))
        job2 = Job(add, function_args=(job1.output,))
        return Response(replace=[job1, job2])

    def replace_flow():
        job = Job(add, function_args=(1,))
        flow = Flow([job], output=job.output)
        return Response(replace=flow)

    def replace_flow_multioutput():
        job1 = Job(add, function_args=(1,))
        job2 = Job(add, function_args=(job1.output,))
        flow = Flow([job1, job2], output={"1": job1.output, "2": job2.output})
        return Response(replace=flow)

    def replace_list_flow():
        job1 = Job(add, function_args=(1,))
        job2 = Job(add, function_args=(job1.output,))
        job3 = Job(add, function_args=(5,))
        job4 = Job(add, function_args=(job3.output,))
        flow1 = Flow([job1, job2], output={"1": job1.output, "2": job2.output})
        flow2 = Flow([job3, job4], output={"3": job3.output, "4": job4.output})
        return Response(replace=[flow1, flow2])

    # replace with job
    metadata = {"hi": "I exist"}
    test_job = Job(replace_job, metadata=metadata, output_schema="123")
    response = test_job.run(memory_jobstore)
    assert isinstance(response.replace, Job)
    assert response.replace.index == 2
    assert response.replace.uuid == test_job.uuid
    assert response.replace.metadata == metadata
    assert response.replace.output_schema == "123"

    # replace with list of job
    test_job = Job(replace_list_job, metadata=metadata, output_schema="123")
    response = test_job.run(memory_jobstore)
    assert isinstance(response.replace, Flow)
    assert response.replace.jobs[-1].function == add
    assert len(response.replace.jobs) == 2
    # currently output schema and metadata ignored in this case

    # replace with flow with outputs
    test_job = Job(replace_flow, metadata=metadata, output_schema="123")
    response = test_job.run(memory_jobstore)
    assert isinstance(response.replace, Flow)
    assert response.replace.jobs[-1].index == 2
    assert response.replace.jobs[-1].uuid == test_job.uuid
    assert response.replace.jobs[-1].metadata == metadata
    assert response.replace.jobs[-1].output_schema == "123"
    assert response.replace.output is not None

    # replace with flow with multi outputs
    test_job = Job(replace_flow_multioutput, metadata=metadata, output_schema="123")
    response = test_job.run(memory_jobstore)
    assert isinstance(response.replace, Flow)
    assert response.replace.jobs[-1].index == 2
    assert response.replace.jobs[-1].uuid == test_job.uuid
    assert response.replace.jobs[-1].metadata == metadata
    assert response.replace.jobs[-1].output_schema == "123"
    assert response.replace.output is not None

    # replace with list of flow
    test_job = Job(replace_list_flow, metadata=metadata, output_schema="123")
    response = test_job.run(memory_jobstore)
    assert isinstance(response.replace, Flow)
    assert isinstance(response.replace.jobs[-1], Flow)
    assert len(response.replace.jobs) == 2
    # currently output schema and metadata ignored in this case


def test_job_config(memory_jobstore):
    from jobflow import (
        CURRENT_JOB,
        Job,
        JobConfig,
        OnMissing,
        OutputReference,
        Response,
    )

    def store_exposed():
        return CURRENT_JOB.store is not None

    def reference_resolved(arg):
        return not isinstance(arg, OutputReference)

    def return_arg(arg):
        return arg

    # test expose store
    config = JobConfig(expose_store=False)
    test_job = Job(store_exposed, config=config)
    response = test_job.run(memory_jobstore)
    assert response.output is False

    config = JobConfig(expose_store=True)
    test_job = Job(store_exposed, config=config)
    response = test_job.run(memory_jobstore)
    assert response.output is True

    ref = OutputReference("1234")
    memory_jobstore.update({"uuid": "1234", "index": 0, "output": 5})
    config = JobConfig(resolve_references=False)
    test_job = Job(reference_resolved, function_args=(ref,), config=config)
    response = test_job.run(memory_jobstore)
    assert response.output is False

    config = JobConfig(resolve_references=True)
    test_job = Job(reference_resolved, function_args=(ref,), config=config)
    response = test_job.run(memory_jobstore)
    assert response.output is True

    ref = OutputReference("xyz")
    config = JobConfig(on_missing_references=OnMissing.ERROR)
    test_job = Job(return_arg, function_args=(ref,), config=config)
    with pytest.raises(ValueError):
        test_job.run(memory_jobstore)

    config = JobConfig(on_missing_references=OnMissing.NONE)
    test_job = Job(return_arg, function_args=(ref,), config=config)
    response = test_job.run(memory_jobstore)
    assert response.output is None

    config = JobConfig(on_missing_references=OnMissing.PASS)
    test_job = Job(return_arg, function_args=(ref,), config=config)
    response = test_job.run(memory_jobstore)
    assert isinstance(response.output, OutputReference)

    # test pass manager config
    def replace_job():
        job = Job(add, function_args=(1,))
        return Response(replace=job)

    def replace_list_job():
        job1 = Job(add, function_args=(1,))
        job2 = Job(add, function_args=(job1.output,))
        return Response(replace=[job1, job2])

    def replace_flow():
        from jobflow import Flow

        job = Job(add, function_args=(1,))
        flow = Flow([job], output=job.output)
        return Response(replace=flow)

    def addition_job():
        job = Job(add, function_args=(1,))
        return Response(addition=job)

    def detour_job():
        job = Job(add, function_args=(1,))
        return Response(detour=job)

    manager_config = {"abc": 1}
    pass_config = JobConfig(manager_config=manager_config, pass_manager_config=True)
    nopass_config = JobConfig(manager_config=manager_config, pass_manager_config=False)

    # test replace
    test_job = Job(replace_job, config=nopass_config)
    response = test_job.run(memory_jobstore)
    assert response.replace.config.manager_config == {}

    test_job = Job(replace_job, config=pass_config)
    response = test_job.run(memory_jobstore)
    assert response.replace.config.manager_config == manager_config

    # test replace list of jobs
    test_job = Job(replace_list_job, config=nopass_config)
    response = test_job.run(memory_jobstore)
    for j in response.replace.jobs:
        assert j.config.manager_config == {}

    test_job = Job(replace_list_job, config=pass_config)
    response = test_job.run(memory_jobstore)
    for j in response.replace.jobs:
        assert j.config.manager_config == manager_config

    # test replace with flow
    test_job = Job(replace_flow, config=nopass_config)
    response = test_job.run(memory_jobstore)
    for j in response.replace.jobs:
        assert j.config.manager_config == {}

    test_job = Job(replace_flow, config=pass_config)
    response = test_job.run(memory_jobstore)
    for j in response.replace.jobs:
        assert j.config.manager_config == manager_config

    # test addition
    test_job = Job(addition_job, config=nopass_config)
    response = test_job.run(memory_jobstore)
    assert response.addition.config.manager_config == {}

    test_job = Job(addition_job, config=pass_config)
    response = test_job.run(memory_jobstore)
    assert response.addition.config.manager_config == manager_config

    # test detour
    test_job = Job(detour_job, config=nopass_config)
    response = test_job.run(memory_jobstore)
    assert response.detour.config.manager_config == {}

    test_job = Job(detour_job, config=pass_config)
    response = test_job.run(memory_jobstore)
    assert response.detour.config.manager_config == manager_config


def test_job_input_references():
    from jobflow.core.job import Job
    from jobflow.core.reference import OutputReference

    ref1 = OutputReference("12345")
    ref2 = OutputReference("54321")
    test_job = Job(add, function_args=(ref1,), function_kwargs={"b": ref2})
    assert set(test_job.input_references) == {ref1, ref2}
    assert set(test_job.input_uuids) == {"12345", "54321"}
    assert test_job.input_references_grouped == {"12345": (ref1,), "54321": (ref2,)}

    # test using reference attributes
    ref1 = OutputReference("12345", attributes=(("i", "name"),))
    ref2 = OutputReference("12345", attributes=(("i", "value"),))
    test_job = Job(add, function_args=(ref1,), function_kwargs={"b": ref2})
    assert set(test_job.input_references) == {ref1, ref2}
    assert set(test_job.input_uuids) == {"12345"}
    assert set(list(test_job.input_references_grouped)) == {"12345"}
    assert set(list(test_job.input_references_grouped["12345"])) == {ref1, ref2}


def test_job_resolve_args(memory_jobstore):
    from jobflow.core.job import Job
    from jobflow.core.reference import OutputReference

    # test basic run with no references
    test_job = Job(print, function_args=("I am a job",))
    resolved_job = test_job.resolve_args(memory_jobstore)
    assert test_job == resolved_job

    ref = OutputReference("1234")
    memory_jobstore.update({"uuid": "1234", "index": 1, "output": 2})

    # test run with input references
    test_job = Job(add, function_args=(1,), function_kwargs={"b": ref})
    resolved_job = test_job.resolve_args(memory_jobstore)
    assert test_job == resolved_job
    assert resolved_job.function_kwargs["b"] == 2

    # test resolve with inplace=False
    test_job = Job(add, function_args=(1,), function_kwargs={"b": ref})
    resolved_job = test_job.resolve_args(memory_jobstore, inplace=False)
    assert test_job != resolved_job
    assert resolved_job.function_kwargs["b"] == 2
    assert isinstance(test_job.function_kwargs["b"], OutputReference)


def test_job_decorator():
    from jobflow.core.job import job

    # test basic init
    decorated = job(print)
    test_job = decorated("I am a job")
    assert test_job
    assert test_job.function == print
    assert test_job.function_args == ("I am a job",)
    assert test_job.function_kwargs == {}
    assert test_job.name == "print"
    assert type(test_job.output).__name__ == "OutputReference"
    assert test_job.uuid is not None
    assert test_job.uuid == test_job.output.uuid

    # test init no args
    decorated = job(print)
    test_job = decorated()
    assert test_job
    assert test_job.function == print
    assert test_job.function_args == tuple()
    assert test_job.function_kwargs == {}
    assert test_job.name == "print"
    assert type(test_job.output).__name__ == "OutputReference"
    assert test_job.uuid is not None
    assert test_job.uuid == test_job.output.uuid

    # test init with outputs
    decorated = job(add)
    test_job = decorated(1, b=2)
    assert test_job
    assert test_job.function == add
    assert test_job.function_args == (1,)
    assert test_job.function_kwargs == {"b": 2}
    assert test_job.name == "add"
    assert test_job.uuid is not None
    assert type(test_job.output).__name__ == "OutputReference"
    assert test_job.uuid == test_job.output.uuid

    # test applying the decorator without arguments
    @job
    def print_message(message):
        print(message)

    test_job = print_message("I am a job")
    assert test_job
    assert test_job.function.__name__ == print_message.__name__
    assert test_job.function.__module__ == print_message.__module__
    assert test_job.function_args == ("I am a job",)
    assert test_job.function_kwargs == {}
    assert test_job.uuid is not None
    assert type(test_job.output).__name__ == "OutputReference"
    assert test_job.uuid == test_job.output.uuid

    # test applying the decorator with arguments
    @job
    def add_numbers(a, b=10):
        return a + b

    test_job = add_numbers(1, b=2)
    assert test_job
    assert test_job.function.__name__ == add_numbers.__name__
    assert test_job.function.__module__ == add_numbers.__module__
    assert test_job.function_args == (1,)
    assert test_job.function_kwargs == {"b": 2}
    assert test_job.uuid is not None
    assert type(test_job.output).__name__ == "OutputReference"
    assert test_job.uuid == test_job.output.uuid


def test_response():
    # no need to test init as it is just a dataclass, instead test from_job_returns
    # test no job returns
    from pydantic import BaseModel

    from jobflow.core.job import Response

    response = Response.from_job_returns(None)
    assert response == Response()

    # test single output
    response = Response.from_job_returns(1)
    assert response.output == 1

    # test list output
    response = Response.from_job_returns([1, 2, 3])
    assert response.output == [1, 2, 3]

    # test tuple output
    response = Response.from_job_returns((1, 2, 3))
    assert response.output == (1, 2, 3)

    # test dict output
    response = Response.from_job_returns({"a": 1, "b": 2})
    assert response.output == {"a": 1, "b": 2}

    # test Response output
    response_original = Response(output=(1, 2, 3), stop_children=True)
    response_processed = Response.from_job_returns(response_original)
    assert response_original == response_processed

    # test Response and another output
    with pytest.raises(ValueError):
        Response.from_job_returns([response_original, 5])

    # test schema
    class MySchema(BaseModel):
        number: int
        name: str

    response = Response.from_job_returns(
        {"number": "5", "name": "Ian"}, output_schema=MySchema
    )
    assert response.output.__class__.__name__ == "MySchema"
    assert response.output.number == 5
    assert response.output.name == "Ian"

    # test schema does not match output
    from pydantic import ValidationError

    with pytest.raises(ValidationError):
        Response.from_job_returns({"number": "5"}, output_schema=MySchema)


def test_serialization():
    import json

    from monty.json import MontyDecoder, MontyEncoder

    from jobflow import Job

    test_job = Job(function=add, function_args=(1,), function_kwargs={"b": 2})

    uuid = test_job.uuid

    encoded_job = json.loads(MontyEncoder().encode(test_job))
    decoded_job = MontyDecoder().process_decoded(encoded_job)

    assert decoded_job.uuid == uuid


def test_class_jobs(memory_jobstore):
    from jobflow.core.job import job

    class Test:
        @job
        @staticmethod
        def static_before(x, y):
            return x + y

        @staticmethod
        @job
        def static_after(x, y):
            return x + y

        @job
        @classmethod
        def class_before(cls, x, y):
            return x + y

        @classmethod
        @job
        def class_after(cls, x, y):
            return x + y

        @job
        def instance(self, x, y):
            return x + y

    # test staticmethod decorator put after job decorator
    test_job = Test.static_after(3, 4)
    response = test_job.run(memory_jobstore)
    assert response.output == 7

    # test staticmethod decorator put before job decorator
    test_job = Test.static_before(3, 4)
    response = test_job.run(memory_jobstore)
    assert response.output == 7

    # test classmethod decorator put after job decorator
    test_job = Test.class_after(3, 4)
    response = test_job.run(memory_jobstore)
    assert response.output == 7

    # test classmethod decorator put before job decorator
    test_job = Test.class_before(3, 4)
    response = test_job.run(memory_jobstore)
    assert response.output == 7

    # test job decorator in instance method
    test_job = Test().instance(3, 4)
    response = test_job.run(memory_jobstore)
    assert response.output == 7


def test_maker():
    from dataclasses import dataclass

    from jobflow import Maker, job

    @dataclass
    class AddMaker(Maker):
        name: str = "add"
        b: int = 2

        @job
        def make(self, a):
            return a + self.b

    maker = AddMaker()
    test_job = maker.make(2)
    assert test_job.maker == maker

    # test setting job name; should also update the maker name
    test_job.name = "abc"
    assert test_job.name == "abc"
    assert test_job.maker.name == "abc"

    test_job.name += "xyz"
    assert test_job.name == "abcxyz"
    assert test_job.maker.name == "abcxyz"


def test_graph():
    from jobflow import Job

    # test no inputs
    test_job = Job(add, function_args=(1,), function_kwargs={"b": 2})
    graph = test_job.graph
    assert len(graph.nodes) == 1
    assert len(graph.edges) == 0

    # test arg inputs
    test_job1 = Job(add, function_args=(1,), function_kwargs={"b": 2})
    test_job = Job(add, function_args=(test_job1.output,), function_kwargs={"b": 2})
    graph = test_job.graph
    assert len(graph.nodes) == 2
    assert len(graph.edges) == 1

    # test arg input properties
    test_job1 = Job(add, function_args=(1,), function_kwargs={"b": 2})
    test_job = Job(
        add, function_args=(test_job1.output.value,), function_kwargs={"b": 2}
    )
    graph = test_job.graph
    assert len(graph.nodes) == 2
    assert len(graph.edges) == 1
    assert graph.get_edge_data(test_job1.uuid, test_job.uuid)["properties"] == "value"

    # test arg and kwargs inputs
    test_job1 = Job(add, function_args=(1,), function_kwargs={"b": 2})
    test_job2 = Job(add, function_args=(1,), function_kwargs={"b": 2})
    test_job = Job(
        add, function_args=(test_job1.output,), function_kwargs={"b": test_job2.output}
    )
    graph = test_job.graph
    assert len(graph.nodes) == 3
    assert len(graph.edges) == 2

    # test arg and kwargs input properties
    test_job1 = Job(add, function_args=(1,), function_kwargs={"b": 2})
    test_job = Job(
        add,
        function_args=(test_job1.output.value,),
        function_kwargs={"b": test_job1.output.name},
    )
    graph = test_job.graph
    assert len(graph.nodes) == 2
    assert len(graph.edges) == 1
    assert set(graph.get_edge_data(test_job1.uuid, test_job.uuid)["properties"]) == {
        "value",
        "name",
    }


def test_set_uuid():
    from jobflow import Job

    test_job = Job(add, function_args=(1,))
    test_job.set_uuid("1234")
    assert test_job.uuid == "1234"
    assert test_job.output.uuid == "1234"


def test_update_kwargs():
    from jobflow import Job

    # test no filter
    test_job = Job(add, function_args=(1,), function_kwargs={"b": 2})
    test_job.update_kwargs({"b": 5})
    assert test_job.function_kwargs["b"] == 5

    # test name filter
    test_job = Job(add, function_args=(1,), function_kwargs={"b": 2})
    test_job.update_kwargs({"b": 5}, name_filter="add")
    assert test_job.function_kwargs["b"] == 5

    test_job = Job(add, function_args=(1,), function_kwargs={"b": 2})
    test_job.update_kwargs({"b": 5}, name_filter="div")
    assert test_job.function_kwargs["b"] == 2

    # test function filter
    test_job = Job(add, function_args=(1,), function_kwargs={"b": 2})
    test_job.update_kwargs({"b": 5}, function_filter=add)
    assert test_job.function_kwargs["b"] == 5

    test_job = Job(add, function_args=(1,), function_kwargs={"b": 2})
    test_job.update_kwargs({"b": 5}, function_filter=list)
    assert test_job.function_kwargs["b"] == 2

    # test dict mod
    test_job = Job(add, function_args=(1,), function_kwargs={"b": 2})
    test_job.update_kwargs({"_inc": {"b": 5}}, dict_mod=True)
    assert test_job.function_kwargs["b"] == 7


def test_update_maker_kwargs():
    from dataclasses import dataclass

    from jobflow import Job, Maker, job

    @dataclass
    class AddMaker(Maker):
        name: str = "add"
        b: int = 2

        @job
        def make(self, a):
            return a + self.b

    @job
    def outer_job(maker1, maker2=None):
        pass

    # test no maker
    test_job = Job(add, function_args=(1,), function_kwargs={"b": 2})
    test_job.update_maker_kwargs({"b": 5})
    assert test_job.function_kwargs["b"] == 2

    add_maker = AddMaker(b=3)

    # test no filter
    test_job = add_maker.make(5)
    test_job.update_maker_kwargs({"b": 10})
    assert test_job.maker.b == 10

    # test bad kwarg
    test_job = add_maker.make(5)
    with pytest.raises(TypeError):
        test_job.update_maker_kwargs({"c": 10})

    # test name filter
    test_job = add_maker.make(5)
    test_job.update_maker_kwargs({"b": 10}, name_filter="add")
    assert test_job.maker.b == 10

    test_job = add_maker.make(5)
    test_job.update_maker_kwargs({"b": 10}, name_filter="div")
    assert test_job.maker.b == 3

    # test class filter
    test_job = add_maker.make(5)
    test_job.update_maker_kwargs({"b": 10}, class_filter=AddMaker)
    assert test_job.maker.b == 10

    test_job = add_maker.make(5)
    test_job.update_maker_kwargs({"b": 10}, class_filter=list)
    assert test_job.maker.b == 3

    # test class filter with instance
    test_job = add_maker.make(5)
    test_job.update_maker_kwargs({"b": 10}, class_filter=add_maker)
    assert test_job.maker.b == 10

    # test dict mod
    test_job = add_maker.make(5)
    test_job.update_maker_kwargs({"_inc": {"b": 10}}, dict_mod=True)
    assert test_job.maker.b == 13

    # test makers in job args and kwargs
    maker1 = AddMaker(b=1)
    maker2 = AddMaker(b=2)
    test_job = outer_job(maker1, maker2=maker2)
    test_job.update_maker_kwargs({"_inc": {"b": 10}}, dict_mod=True)
    assert test_job.function_args[0].b == 11
    assert test_job.function_kwargs["maker2"].b == 12

    # test nested=False does not impact the jobs
    test_job.update_maker_kwargs({"_inc": {"b": 20}}, dict_mod=True, nested=False)
    assert test_job.function_args[0].b == 11
    assert test_job.function_kwargs["maker2"].b == 12


def test_append_name():
    from jobflow import Job

    # test append
    test_job = Job(function=add)
    test_job.append_name(" test")
    assert test_job.name == "add test"

    # test prepend
    test_job = Job(function=add)
    test_job.append_name("test ", prepend=True)
    assert test_job.name == "test add"


def test_output_schema(memory_jobstore):
    from pydantic import BaseModel

    from jobflow import Job, Response, job

    class AddSchema(BaseModel):
        result: int

    @job(output_schema=AddSchema)
    def add_schema(a, b):
        return AddSchema(**{"result": a + b})

    @job(output_schema=AddSchema)
    def add_schema_dict(a, b):
        return {"result": a + b}

    @job(output_schema=AddSchema)
    def add_schema_bad(a, b):
        return a + b

    @job(output_schema=AddSchema)
    def add_schema_wrong_key(a, b):
        return {"bad_key": a + b}

    @job(output_schema=AddSchema)
    def add_schema_no_output(a, b):
        a + b
        return None

    @job(output_schema=AddSchema)
    def add_schema_response_dict(a, b):
        return Response(output={"result": a + b})

    @job(output_schema=AddSchema)
    def add_schema_response(a, b):
        return Response(output=AddSchema(result=a + b))

    @job
    def add_schema_replace(a, b):
        new_job = Job(add, function_args=(a, b), output_schema=AddSchema)
        return Response(replace=new_job)

    test_job = add_schema(5, 6)
    response = test_job.run(memory_jobstore)
    assert response.output.__class__.__name__ == "AddSchema"
    assert response.output.result == 11

    test_job = add_schema_dict(5, 6)
    response = test_job.run(memory_jobstore)
    assert response.output.__class__.__name__ == "AddSchema"
    assert response.output.result == 11

    test_job = add_schema_response(5, 6)
    response = test_job.run(memory_jobstore)
    assert response.output.__class__.__name__ == "AddSchema"
    assert response.output.result == 11

    test_job = add_schema_response_dict(5, 6)
    response = test_job.run(memory_jobstore)
    assert response.output.__class__.__name__ == "AddSchema"
    assert response.output.result == 11

    test_job = add_schema_replace(5, 6)
    response = test_job.run(memory_jobstore)
    assert response.replace.output_schema.__name__ == "AddSchema"

    test_job = add_schema_bad(5, 6)
    with pytest.raises(ValueError):
        test_job.run(memory_jobstore)

    test_job = add_schema_wrong_key(5, 6)
    with pytest.raises(ValueError):
        test_job.run(memory_jobstore)

    test_job = add_schema_no_output(5, 6)
    with pytest.raises(ValueError):
        test_job.run(memory_jobstore)


def test_store_inputs(memory_jobstore):
    from jobflow.core.job import OutputReference, store_inputs

    test_job = store_inputs(1)
    test_job.run(memory_jobstore)
    output = memory_jobstore.query_one({"uuid": test_job.uuid}, ["output"])["output"]
    assert output == 1

    ref = OutputReference("abcd")
    test_job = store_inputs(ref)
    test_job.run(memory_jobstore)
    output = memory_jobstore.query_one({"uuid": test_job.uuid}, ["output"])["output"]
    assert OutputReference.from_dict(output) == ref


def test_pass_manager_config():
    from jobflow import Flow, Job
    from jobflow.core.job import pass_manager_config

    manager_config = {"abc": 1}

    # test single job
    test_job1 = Job(add, function_args=(1,))
    pass_manager_config(test_job1, manager_config)
    assert test_job1.config.manager_config == manager_config

    # test list job
    test_job1 = Job(add, function_args=(1,))
    test_job2 = Job(add, function_args=(1,))
    pass_manager_config([test_job1, test_job2], manager_config)
    assert test_job1.config.manager_config == manager_config
    assert test_job2.config.manager_config == manager_config

    # test flow
    test_job1 = Job(add, function_args=(1,))
    test_job2 = Job(add, function_args=(1,))
    flow = Flow([test_job1, test_job2])
    pass_manager_config(flow, manager_config)
    assert test_job1.config.manager_config == manager_config
    assert test_job2.config.manager_config == manager_config

    # test bad input
    with pytest.raises(ValueError):
        pass_manager_config(["str"], manager_config)



"""This module provides helper fixtures for quickly generating flows and jobs."""

import pytest


@pytest.fixture(scope="session")
def simple_job():
    from jobflow import job

    global func

    @job
    def func(message):
        return message + "_end"

    return func


@pytest.fixture(scope="session")
def simple_flow(simple_job):
    from jobflow import Flow

    def _gen():
        simple = simple_job("12345")
        return Flow([simple], simple.output)

    return _gen


@pytest.fixture(scope="session")
def connected_flow(simple_job):
    from jobflow import Flow

    def _gen():
        simple1 = simple_job("12345")
        simple2 = simple_job(simple1.output)
        return Flow([simple1, simple2], simple2.output, "Connected Flow")

    return _gen


@pytest.fixture(scope="session")
def nested_flow(connected_flow):
    from jobflow import Flow

    def _gen():
        flow1 = connected_flow()
        flow2 = connected_flow()
        flow2.jobs[0].function_args = (flow1.jobs[1].output,)
        return Flow([flow1, flow2], flow2.output)

    return _gen


@pytest.fixture(scope="session")
def addition_job(simple_job):
    from jobflow import Response, job

    global addition_func

    @job
    def addition_func(a, b):
        return Response(output=a + b, addition=simple_job(str(a + b)))

    return addition_func


@pytest.fixture(scope="session")
def addition_flow(addition_job):
    from jobflow import Flow

    def _gen():
        add = addition_job(5, 6)
        return Flow([add], add.output)

    return _gen


@pytest.fixture(scope="session")
def detour_job(simple_job):
    from jobflow import Response, job

    global detour_func

    @job
    def detour_func(a, b):
        return Response(output=a + b, detour=simple_job(str(a + b)))

    return detour_func


@pytest.fixture(scope="session")
def detour_flow(detour_job, simple_job):
    from jobflow import Flow, JobOrder

    def _gen():
        detour = detour_job(5, 6)
        simple = simple_job("12345")
        return Flow([detour, simple], simple.output, order=JobOrder.LINEAR)

    return _gen


@pytest.fixture(scope="session")
def replace_job(simple_job):
    from jobflow import Response, job

    global replace_func

    @job
    def replace_func(a, b):
        return Response(output=a + b, replace=simple_job(str(a + b)))

    return replace_func


@pytest.fixture(scope="session")
def replace_job_with_flow(simple_job):
    from jobflow import Flow, Response, job

    global replace_func_flow

    @job
    def replace_func_flow(a, b):
        first_job = simple_job(str(a + b))
        second_job = simple_job(first_job.output)
        flow = Flow(
            [first_job, second_job],
            {"first": first_job.output, "second": second_job.output},
        )
        return Response(output=a + b, replace=flow)

    return replace_func_flow


@pytest.fixture(scope="session")
def replace_flow(replace_job, simple_job):
    from jobflow import Flow, JobOrder

    def _gen():
        replace = replace_job(5, 6)
        simple = simple_job("12345")
        return Flow([replace, simple], simple.output, order=JobOrder.LINEAR)

    return _gen


@pytest.fixture(scope="session")
def replace_flow_nested(replace_job_with_flow, simple_job):
    from jobflow import Flow, JobOrder

    def _gen():
        replace = replace_job_with_flow(5, 6)
        simple = simple_job("12345")
        return Flow([replace, simple], simple.output, order=JobOrder.LINEAR)

    return _gen


@pytest.fixture(scope="session")
def stop_jobflow_job():
    from jobflow import Response, job

    global stop_jobflow_func

    @job
    def stop_jobflow_func():
        return Response(output="1234", stop_jobflow=True)

    return stop_jobflow_func


@pytest.fixture(scope="session")
def stop_jobflow_flow(stop_jobflow_job, simple_job):
    from jobflow import Flow, JobOrder

    def _gen():
        stop = stop_jobflow_job()
        simple = simple_job("12345")
        return Flow([stop, simple], simple.output, order=JobOrder.LINEAR)

    return _gen


@pytest.fixture(scope="session")
def stop_children_job():
    from jobflow import Response, job

    global stop_children_func

    @job
    def stop_children_func():
        return Response(output="1234", stop_children=True)

    return stop_children_func


@pytest.fixture(scope="session")
def stop_children_flow(stop_children_job, simple_job):
    from jobflow import Flow

    def _gen():
        stop = stop_children_job()
        simple1 = simple_job(stop.output)
        simple2 = simple_job("12345")
        return Flow([stop, simple1, simple2], simple2.output)

    return _gen


@pytest.fixture(scope="session")
def error_job():
    from jobflow import job

    global error_func

    @job
    def error_func():
        raise ValueError("errored")

    return error_func


@pytest.fixture(scope="session")
def error_flow(error_job, simple_job):
    from jobflow import Flow

    def _gen():
        error = error_job()
        simple1 = simple_job(error.output)
        simple2 = simple_job(simple1.output)
        return Flow([error, simple1, simple2])

    return _gen


@pytest.fixture(scope="session")
def stored_data_job():
    from jobflow import Response, job

    global stored_data_func

    @job
    def stored_data_func(message):
        return Response(output=message + "_end", stored_data={"a": "message"})

    return stored_data_func


@pytest.fixture(scope="session")
def stored_data_flow(stored_data_job):
    from jobflow import Flow

    def _gen():
        store = stored_data_job("12345")
        return Flow([store])

    return _gen


@pytest.fixture(scope="session")
def detour_stop_job(stop_jobflow_job):
    from jobflow import Response, job

    global detour_stop_func

    @job
    def detour_stop_func(a, b):
        return Response(output=a + b, detour=stop_jobflow_job())

    return detour_stop_func


@pytest.fixture(scope="session")
def detour_stop_flow(detour_stop_job, simple_job):
    from jobflow import Flow, JobOrder

    def _gen():
        detour = detour_stop_job(5, 6)
        simple = simple_job("12345")
        return Flow([detour, simple], simple.output, order=JobOrder.LINEAR)

    return _gen


@pytest.fixture(scope="session")
def fw_dir():
    # longer session
    import os
    import shutil
    import tempfile

    old_cwd = os.getcwd()
    newpath = tempfile.mkdtemp()
    os.chdir(newpath)

    yield

    os.chdir(old_cwd)
    shutil.rmtree(newpath)


@pytest.fixture(scope="session")
def replace_and_detour_job(simple_job):
    from jobflow import Response, job

    global replace_and_detour_func

    @job
    def replace_and_detour_func(a, b):
        return Response(
            output=a + b, replace=simple_job(str(a + b)), detour=simple_job("xyz")
        )

    return replace_and_detour_func


@pytest.fixture(scope="session")
def replace_and_detour_flow(replace_and_detour_job, simple_job):
    from jobflow import Flow, JobOrder

    def _gen():
        replace = replace_and_detour_job(5, 6)
        simple = simple_job("12345")
        return Flow([replace, simple], simple.output, order=JobOrder.LINEAR)

    return _gen



import pytest


def test_flow_to_workflow(
    memory_jobstore, simple_job, simple_flow, connected_flow, nested_flow
):
    from fireworks import Workflow

    from jobflow import OnMissing
    from jobflow.managers.fireworks import flow_to_workflow

    # test simple job
    flow = simple_job()
    wf = flow_to_workflow(flow, memory_jobstore)

    assert type(wf) == Workflow
    assert wf.name == "Flow"
    assert len(wf.fws) == 1
    assert wf.fws[0].name == "func"

    # test simple job no store
    flow = simple_job()
    wf = flow_to_workflow(flow)

    assert type(wf) == Workflow
    assert wf.name == "Flow"
    assert len(wf.fws) == 1
    assert wf.fws[0].name == "func"

    # test simple flow
    flow = simple_flow()
    wf = flow_to_workflow(flow, memory_jobstore)

    assert type(wf) == Workflow
    assert wf.name == "Flow"
    assert len(wf.fws) == 1
    assert wf.fws[0].name == "func"

    # test connected flow
    flow = connected_flow()
    wf = flow_to_workflow(flow, memory_jobstore)

    assert type(wf) == Workflow
    assert wf.name == "Connected Flow"
    assert len(wf.fws) == 2
    assert wf.fws[0].name == "func"
    assert wf.fws[1].name == "func"

    # test nested flow
    flow = nested_flow()
    wf = flow_to_workflow(flow, memory_jobstore)

    assert len(wf.fws) == 4
    assert wf.fws[0].name == "func"
    assert wf.fws[1].name == "func"
    assert wf.fws[2].name == "func"
    assert wf.fws[3].name == "func"

    # test on missing causes allow fizzled parents
    flow = connected_flow()
    flow.jobs[0].config.on_missing_references = OnMissing.NONE
    wf = flow_to_workflow(flow, memory_jobstore)

    assert wf.fws[0].spec["_allow_fizzled_parents"] is True

    # test manager config
    flow = connected_flow()
    flow.jobs[0].config.manager_config = {"metadata": 5}
    wf = flow_to_workflow(flow, memory_jobstore)
    assert wf.fws[0].spec["metadata"] == 5


def test_job_to_firework(
    memory_jobstore, simple_job, simple_flow, connected_flow, nested_flow
):
    from fireworks import Firework

    from jobflow.managers.fireworks import job_to_firework

    # test simple job
    job = simple_job()
    fw = job_to_firework(job, memory_jobstore)

    assert type(fw) == Firework
    assert fw.name == "func"

    job2 = simple_job()
    fw = job_to_firework(
        job2, memory_jobstore, parents=[job.uuid], parent_mapping={job.uuid: 1}
    )

    assert type(fw) == Firework
    assert fw.name == "func"

    with pytest.raises(ValueError):
        job_to_firework(job2, memory_jobstore, parents=[job.uuid])


def test_simple_flow(lpad, mongo_jobstore, fw_dir, simple_flow, capsys):
    from fireworks.core.rocket_launcher import rapidfire

    from jobflow.managers.fireworks import flow_to_workflow

    flow = simple_flow()
    uuid = flow.jobs[0].uuid

    wf = flow_to_workflow(flow, mongo_jobstore)
    fw_ids = lpad.add_wf(wf)

    # run the workflow
    rapidfire(lpad)

    # check workflow completed
    fw_id = list(fw_ids.values())[0]
    wf = lpad.get_wf_by_fw_id(fw_id)

    assert all([s == "COMPLETED" for s in wf.fw_states.values()])

    # check store has the activity output
    result = mongo_jobstore.query_one({"uuid": uuid})
    assert result["output"] == "12345_end"

    # check logs printed
    captured = capsys.readouterr()
    assert "INFO Starting job - func" in captured.out
    assert "INFO Finished job - func" in captured.out


def test_simple_flow_no_store(lpad, fw_dir, simple_flow, capsys):
    from fireworks.core.rocket_launcher import rapidfire

    from jobflow import SETTINGS
    from jobflow.managers.fireworks import flow_to_workflow

    flow = simple_flow()
    uuid = flow.jobs[0].uuid

    wf = flow_to_workflow(flow)
    fw_ids = lpad.add_wf(wf)

    # run the workflow
    rapidfire(lpad)

    # check workflow completed
    fw_id = list(fw_ids.values())[0]
    wf = lpad.get_wf_by_fw_id(fw_id)

    assert all([s == "COMPLETED" for s in wf.fw_states.values()])

    # check store has the activity output
    result = SETTINGS.JOB_STORE.query_one({"uuid": uuid})
    assert result["output"] == "12345_end"

    # check logs printed
    captured = capsys.readouterr()
    assert "INFO Starting job - func" in captured.out
    assert "INFO Finished job - func" in captured.out


def test_simple_flow_metadata(lpad, mongo_jobstore, fw_dir, simple_flow, capsys):
    from fireworks.core.rocket_launcher import rapidfire

    from jobflow.managers.fireworks import flow_to_workflow

    flow = simple_flow()
    uuid = flow.jobs[0].uuid
    flow.jobs[0].metadata = {"tags": ["my_flow"]}

    wf = flow_to_workflow(flow, mongo_jobstore)
    fw_ids = lpad.add_wf(wf)

    # run the workflow
    rapidfire(lpad)

    # check workflow completed
    fw_id = list(fw_ids.values())[0]
    wf = lpad.get_wf_by_fw_id(fw_id)

    assert all([s == "COMPLETED" for s in wf.fw_states.values()])
    assert wf.fws[0].spec["tags"] == ["my_flow"]

    # check store has the activity output
    result = mongo_jobstore.query_one({"uuid": uuid})
    assert result["output"] == "12345_end"
    assert result["metadata"]["fw_id"] == fw_id

    # test override
    flow = simple_flow()
    flow.jobs[0].config.manager_config = {"_add_launchpad_and_fw_id": False}
    uuid = flow.jobs[0].uuid

    wf = flow_to_workflow(flow, mongo_jobstore)
    lpad.add_wf(wf)

    # run the workflow
    rapidfire(lpad)

    result = mongo_jobstore.query_one({"uuid": uuid})
    assert result["metadata"] == {}


def test_connected_flow(lpad, mongo_jobstore, fw_dir, connected_flow, capsys):
    from fireworks.core.rocket_launcher import rapidfire

    from jobflow.managers.fireworks import flow_to_workflow

    flow = connected_flow()
    uuid1 = flow.jobs[0].uuid
    uuid2 = flow.jobs[1].uuid

    wf = flow_to_workflow(flow, mongo_jobstore)
    fw_ids = lpad.add_wf(wf)

    # run the workflow
    rapidfire(lpad)

    # check workflow completed
    fw_id = list(fw_ids.values())[0]
    wf = lpad.get_wf_by_fw_id(fw_id)

    assert all([s == "COMPLETED" for s in wf.fw_states.values()])

    # check store has the activity output
    result1 = mongo_jobstore.query_one({"uuid": uuid1})
    result2 = mongo_jobstore.query_one({"uuid": uuid2})

    assert result1["output"] == "12345_end"
    assert result2["output"] == "12345_end_end"


def test_nested_flow(lpad, mongo_jobstore, fw_dir, nested_flow, capsys):
    from fireworks.core.rocket_launcher import rapidfire

    from jobflow.managers.fireworks import flow_to_workflow

    flow = nested_flow()
    uuid1 = flow.jobs[0].jobs[0].uuid
    uuid2 = flow.jobs[0].jobs[1].uuid
    uuid3 = flow.jobs[1].jobs[0].uuid
    uuid4 = flow.jobs[1].jobs[1].uuid

    wf = flow_to_workflow(flow, mongo_jobstore)
    fw_ids = lpad.add_wf(wf)

    # run the workflow
    rapidfire(lpad)

    # check workflow completed
    fw_id = list(fw_ids.values())[0]
    wf = lpad.get_wf_by_fw_id(fw_id)

    assert all([s == "COMPLETED" for s in wf.fw_states.values()])

    # check store has the activity output
    result1 = mongo_jobstore.query_one({"uuid": uuid1})
    result2 = mongo_jobstore.query_one({"uuid": uuid2})
    result3 = mongo_jobstore.query_one({"uuid": uuid3})
    result4 = mongo_jobstore.query_one({"uuid": uuid4})

    assert result1["output"] == "12345_end"
    assert result2["output"] == "12345_end_end"
    assert result3["output"] == "12345_end_end_end"
    assert result4["output"] == "12345_end_end_end_end"


def test_addition_flow(lpad, mongo_jobstore, fw_dir, addition_flow, capsys):
    from fireworks.core.rocket_launcher import rapidfire

    from jobflow.managers.fireworks import flow_to_workflow

    flow = addition_flow()
    uuid1 = flow.jobs[0].uuid

    wf = flow_to_workflow(flow, mongo_jobstore)
    fw_ids = lpad.add_wf(wf)

    # run the workflow
    rapidfire(lpad)

    # check workflow completed
    fw_id = list(fw_ids.values())[0]
    wf = lpad.get_wf_by_fw_id(fw_id)

    uuids = [fw.tasks[0]["job"].uuid for fw in wf.fws]
    uuid2 = [u for u in uuids if u != uuid1][0]
    assert all([s == "COMPLETED" for s in wf.fw_states.values()])

    # check store has the activity output
    result1 = mongo_jobstore.query_one({"uuid": uuid1})
    result2 = mongo_jobstore.query_one({"uuid": uuid2})

    assert result1["output"] == 11
    assert result2["output"] == "11_end"


def test_detour_flow(lpad, mongo_jobstore, fw_dir, detour_flow, capsys):
    from fireworks.core.rocket_launcher import rapidfire

    from jobflow.managers.fireworks import flow_to_workflow

    flow = detour_flow()
    uuid1 = flow.jobs[0].uuid
    uuid3 = flow.jobs[1].uuid

    wf = flow_to_workflow(flow, mongo_jobstore)
    fw_ids = lpad.add_wf(wf)

    # run the workflow
    rapidfire(lpad)

    # check workflow completed
    fw_id = list(fw_ids.values())[0]
    wf = lpad.get_wf_by_fw_id(fw_id)

    uuids = [fw.tasks[0]["job"].uuid for fw in wf.fws]
    uuid2 = [u for u in uuids if u != uuid1 and u != uuid3][0]
    assert all([s == "COMPLETED" for s in wf.fw_states.values()])

    # check store has the activity output
    result1 = mongo_jobstore.query_one({"uuid": uuid1})
    result2 = mongo_jobstore.query_one({"uuid": uuid2})
    result3 = mongo_jobstore.query_one({"uuid": uuid3})

    assert result1["output"] == 11
    assert result2["output"] == "11_end"
    assert result3["output"] == "12345_end"

    # assert job2 (detoured job) ran before job3
    assert result2["completed_at"] < result3["completed_at"]


def test_replace_flow(lpad, mongo_jobstore, fw_dir, replace_flow, capsys):
    from fireworks.core.rocket_launcher import rapidfire

    from jobflow.managers.fireworks import flow_to_workflow

    flow = replace_flow()
    uuid1 = flow.jobs[0].uuid
    uuid2 = flow.jobs[1].uuid

    wf = flow_to_workflow(flow, mongo_jobstore)
    fw_ids = lpad.add_wf(wf)

    # run the workflow
    rapidfire(lpad)

    # check workflow completed
    fw_id = list(fw_ids.values())[0]
    wf = lpad.get_wf_by_fw_id(fw_id)

    assert all([s == "COMPLETED" for s in wf.fw_states.values()])

    # check store has the activity output
    result1 = mongo_jobstore.query_one({"uuid": uuid1, "index": 1})
    result2 = mongo_jobstore.query_one({"uuid": uuid1, "index": 2})
    result3 = mongo_jobstore.query_one({"uuid": uuid2, "index": 1})

    assert result1["output"] == 11
    assert result2["output"] == "11_end"
    assert result3["output"] == "12345_end"

    # assert job2 (replaced job) ran before job3
    assert result2["completed_at"] < result3["completed_at"]


def test_stop_jobflow_flow(lpad, mongo_jobstore, fw_dir, stop_jobflow_flow, capsys):
    from fireworks.core.rocket_launcher import rapidfire

    from jobflow.managers.fireworks import flow_to_workflow

    flow = stop_jobflow_flow()
    uuid1 = flow.jobs[0].uuid
    uuid2 = flow.jobs[1].uuid

    wf = flow_to_workflow(flow, mongo_jobstore)
    fw_ids = lpad.add_wf(wf)

    # run the workflow
    rapidfire(lpad)

    # check workflow completed
    fw_id = list(fw_ids.values())[0]
    wf = lpad.get_wf_by_fw_id(fw_id)

    assert set(wf.fw_states.values()) == {"COMPLETED", "DEFUSED"}

    # check store has the activity output
    result1 = mongo_jobstore.query_one({"uuid": uuid1})
    result2 = mongo_jobstore.query_one({"uuid": uuid2})

    assert result1["output"] == "1234"
    assert result2 is None


def test_stop_jobflow_job(lpad, mongo_jobstore, fw_dir, stop_jobflow_job, capsys):
    from fireworks.core.rocket_launcher import rapidfire

    from jobflow.managers.fireworks import flow_to_workflow

    job = stop_jobflow_job()
    uuid1 = job.uuid

    wf = flow_to_workflow(job, mongo_jobstore)
    fw_ids = lpad.add_wf(wf)

    # run the workflow
    rapidfire(lpad)

    # check workflow completed
    fw_id = list(fw_ids.values())[0]
    wf = lpad.get_wf_by_fw_id(fw_id)

    assert list(wf.fw_states.values()) == ["COMPLETED"]

    # check store has the activity output
    result1 = mongo_jobstore.query_one({"uuid": uuid1})

    assert result1["output"] == "1234"


def test_stop_children_flow(lpad, mongo_jobstore, fw_dir, stop_children_flow, capsys):
    from collections import Counter

    from fireworks.core.rocket_launcher import rapidfire

    from jobflow.managers.fireworks import flow_to_workflow

    flow = stop_children_flow()
    uuid1 = flow.jobs[0].uuid
    uuid2 = flow.jobs[1].uuid
    uuid3 = flow.jobs[2].uuid

    wf = flow_to_workflow(flow, mongo_jobstore)
    fw_ids = lpad.add_wf(wf)

    # run the workflow
    rapidfire(lpad)

    # check workflow completed
    fw_id = list(fw_ids.values())[0]
    wf = lpad.get_wf_by_fw_id(fw_id)

    states = Counter(wf.fw_states.values())
    assert states["COMPLETED"] == 2
    assert states["DEFUSED"] == 1

    # check store has the activity output
    result1 = mongo_jobstore.query_one({"uuid": uuid1})
    result2 = mongo_jobstore.query_one({"uuid": uuid2})
    result3 = mongo_jobstore.query_one({"uuid": uuid3})

    assert result1["output"] == "1234"
    assert result2 is None
    assert result3["output"] == "12345_end"


def test_error_flow(lpad, mongo_jobstore, fw_dir, error_flow):
    from fireworks.core.rocket_launcher import rapidfire

    from jobflow.managers.fireworks import flow_to_workflow

    flow = error_flow()
    uuid = flow.jobs[0].uuid

    wf = flow_to_workflow(flow, mongo_jobstore)
    fw_ids = lpad.add_wf(wf)

    # run the workflow
    rapidfire(lpad)
    # check workflow completed
    fw_id = list(fw_ids.values())[0]
    wf = lpad.get_wf_by_fw_id(fw_id)

    assert set(wf.fw_states.values()) == {"WAITING", "FIZZLED"}

    result1 = mongo_jobstore.query_one({"uuid": uuid})

    assert result1 is None


def test_stored_data_flow(lpad, mongo_jobstore, fw_dir, stored_data_flow, capsys):
    from fireworks.core.rocket_launcher import rapidfire

    from jobflow.managers.fireworks import flow_to_workflow

    flow = stored_data_flow()
    flow.jobs[0].uuid

    wf = flow_to_workflow(flow, mongo_jobstore)
    fw_ids = lpad.add_wf(wf)

    # run the workflow
    rapidfire(lpad)

    # check workflow completed
    fw_id = list(fw_ids.values())[0]
    wf = lpad.get_wf_by_fw_id(fw_id)

    assert list(wf.fw_states.values()) == ["COMPLETED"]

    result = lpad.db.launches.find_one({"fw_id": fw_id})
    assert result["action"]["stored_data"] == {"a": "message"}


def test_detour_stop_flow(lpad, mongo_jobstore, fw_dir, detour_stop_flow, capsys):
    from fireworks.core.rocket_launcher import rapidfire

    from jobflow.managers.fireworks import flow_to_workflow

    flow = detour_stop_flow()
    uuid1 = flow.jobs[0].uuid
    uuid3 = flow.jobs[1].uuid

    wf = flow_to_workflow(flow, mongo_jobstore)
    fw_ids = lpad.add_wf(wf)

    # run the workflow
    rapidfire(lpad)

    # check workflow completed
    fw_id = list(fw_ids.values())[0]
    wf = lpad.get_wf_by_fw_id(fw_id)

    uuids = [fw.tasks[0]["job"].uuid for fw in wf.fws]
    uuid2 = [u for u in uuids if u != uuid1 and u != uuid3][0]
    assert list(wf.fw_states.values()) == ["DEFUSED", "COMPLETED", "COMPLETED"]

    # check store has the activity output
    result1 = mongo_jobstore.query_one({"uuid": uuid1})
    result2 = mongo_jobstore.query_one({"uuid": uuid2})
    result3 = mongo_jobstore.query_one({"uuid": uuid3})

    assert result1["output"] == 11
    assert result2["output"] == "1234"
    assert result3 is None


def test_replace_and_detour_flow(
    lpad, mongo_jobstore, fw_dir, replace_and_detour_flow, capsys
):
    from fireworks.core.rocket_launcher import rapidfire

    from jobflow.managers.fireworks import flow_to_workflow

    flow = replace_and_detour_flow()
    uuid1 = flow.jobs[0].uuid
    uuid3 = flow.jobs[1].uuid

    wf = flow_to_workflow(flow, mongo_jobstore)
    fw_ids = lpad.add_wf(wf)

    # run the workflow
    rapidfire(lpad)

    # check workflow completed
    fw_id = list(fw_ids.values())[0]
    wf = lpad.get_wf_by_fw_id(fw_id)

    uuids = [fw.tasks[0]["job"].uuid for fw in wf.fws]
    uuid2 = [u for u in uuids if u != uuid1 and u != uuid3][0]

    assert all([s == "COMPLETED" for s in wf.fw_states.values()])

    # check store has the activity output
    result1 = mongo_jobstore.query_one({"uuid": uuid1, "index": 1})
    result2 = mongo_jobstore.query_one({"uuid": uuid1, "index": 2})
    result3 = mongo_jobstore.query_one({"uuid": uuid2, "index": 1})
    result4 = mongo_jobstore.query_one({"uuid": uuid3, "index": 1})

    assert result1["output"] == 11
    assert result2["output"] == "11_end"
    assert result3["output"] == "xyz_end"
    assert result4["output"] == "12345_end"



def test_initialize_logger(capsys):
    import logging

    from jobflow import initialize_logger

    logger = logging.getLogger("jobflow")

    # initialize logger with default values
    initialize_logger()
    logger.info("123")
    logger.debug("ABC")

    captured = capsys.readouterr()
    assert "INFO 123" in captured.out
    assert "DEBUG" not in captured.out

    # initialize logger with debug level
    initialize_logger(level=logging.DEBUG)
    logger.info("123")
    logger.debug("ABC")

    captured = capsys.readouterr()
    assert "INFO 123" in captured.out
    assert "DEBUG ABC" in captured.out



import pytest


def test_access():
    from jobflow import OutputReference

    # test empty
    ref = OutputReference("123")
    assert ref.attributes == tuple()

    # test bad init
    with pytest.raises(ValueError):
        OutputReference("123", (("x", 1),))

    new_ref = ref.a
    assert new_ref.attributes == (("a", "a"),)
    assert new_ref.uuid == "123"
    assert isinstance(new_ref, OutputReference)

    new_ref = ref["a"]
    assert new_ref.attributes == (("i", "a"),)
    assert new_ref.uuid == "123"
    assert isinstance(new_ref, OutputReference)

    new_ref = ref[1]
    assert new_ref.attributes == (("i", 1),)
    assert new_ref.uuid == "123"
    assert isinstance(new_ref, OutputReference)

    # test filled
    ref = OutputReference("123", (("a", "b"),))

    new_ref = ref.a
    assert new_ref.attributes == (("a", "b"), ("a", "a"))
    assert new_ref.uuid == "123"
    assert isinstance(new_ref, OutputReference)

    new_ref = ref["a"]
    assert new_ref.attributes == (("a", "b"), ("i", "a"))
    assert new_ref.uuid == "123"
    assert isinstance(new_ref, OutputReference)

    with pytest.raises(AttributeError):
        _ = ref.args

    with pytest.raises(AttributeError):
        _ = ref.__fake_variable


def test_get_set_attr():
    from jobflow import OutputReference

    ref = OutputReference("123")

    # these should fail
    with pytest.raises(TypeError):
        ref["a"] = 1

    with pytest.raises(TypeError):
        ref[1] = 1

    with pytest.raises(TypeError):
        ref.a = 1

    ref.uuid = 1
    assert ref.uuid == 1


def test_repr():
    from jobflow import OutputReference

    ref = OutputReference("123")
    assert str(ref) == "OutputReference(123)"

    ref = OutputReference("123", (("a", "a"),))
    assert str(ref) == "OutputReference(123, .a)"

    ref = OutputReference("123", (("a", "a"), ("i", 1)))
    assert str(ref) == "OutputReference(123, .a, [1])"


def test_hash():
    from jobflow import OutputReference

    assert hash(OutputReference("123")) == hash(OutputReference("123"))
    assert hash(OutputReference("123", (("i", 1), ("i", 2)))) == hash(
        OutputReference("123", (("i", 1), ("i", 2)))
    )
    assert hash(OutputReference("123", (("a", "b"), ("i", 2)))) == hash(
        OutputReference("123", (("a", "b"), ("i", 2)))
    )
    assert hash(OutputReference("123", (("a", "b"), ("i", "2")))) == hash(
        OutputReference("123", (("a", "b"), ("i", "2")))
    )


def test_eq():
    from jobflow import OutputReference

    assert OutputReference("123") == OutputReference("123")
    assert OutputReference("123") != OutputReference("1234")
    assert OutputReference("123", (("i", 1),)) == OutputReference("123", (("i", 1),))
    assert OutputReference("123", (("i", "a"),)) == OutputReference(
        "123", (("i", "a"),)
    )
    assert OutputReference("123", (("a", "a"),)) == OutputReference(
        "123", (("a", "a"),)
    )
    assert OutputReference("123", (("a", "a"), ("i", "b"))) == OutputReference(
        "123", (("a", "a"), ("i", "b"))
    )
    assert OutputReference("123", (("i", 1),)) != OutputReference("1234", (("i", 1),))
    assert OutputReference("123", (("i", 1),)) != OutputReference("123", (("i", 2),))
    assert OutputReference("123", (("i", 1),)) != OutputReference(
        "123", (("i", 2), ("i", 3), ("i", 4))
    )
    assert OutputReference("123", (("i", 1),)) != "OutputReference(123, [1])"


def test_as_dict():
    from jobflow import OutputReference

    ref = OutputReference("123")
    d = ref.as_dict()
    assert d["@class"] == "OutputReference"
    assert d["@module"] == "jobflow.core.reference"
    assert d["uuid"] == "123"

    ref = OutputReference("123", (("a", "a"), ("i", "b")))
    d = ref.as_dict()
    assert d["@class"] == "OutputReference"
    assert d["@module"] == "jobflow.core.reference"
    assert d["uuid"] == "123"
    assert d["attributes"] == (("a", "a"), ("i", "b"))


def test_set_uuid():
    from jobflow import OutputReference

    ref = OutputReference("123")
    new_ref = ref.set_uuid("321")
    assert ref.uuid == "321"
    assert new_ref.uuid == "321"

    ref = OutputReference("123")
    new_ref = ref.set_uuid("321", inplace=False)
    assert ref.uuid == "123"
    assert new_ref.uuid == "321"


def test_schema():
    from pydantic import BaseModel

    from jobflow import OutputReference

    class MySchema(BaseModel):
        number: int
        name: str

    ref = OutputReference("123", output_schema=MySchema)
    assert ref.attributes == tuple()

    # check valid schema access works
    new_ref = ref.number
    assert new_ref.uuid == "123"
    assert new_ref.output_schema is None

    new_ref = ref["name"]
    assert new_ref.uuid == "123"
    assert new_ref.output_schema is None

    with pytest.raises(AttributeError):
        assert ref.a.uuid == "123"

    with pytest.raises(AttributeError):
        assert ref["a"].uuid == "123"

    with pytest.raises(AttributeError):
        assert ref[1].uuid == "123"


def test_resolve(memory_jobstore):
    from jobflow import OnMissing, OutputReference

    ref = OutputReference("123")

    # fail if cache or store not provided
    with pytest.raises(TypeError):
        ref.resolve()

    # test on missing
    assert ref.resolve(memory_jobstore, on_missing=OnMissing.NONE) is None
    assert ref.resolve(memory_jobstore, on_missing=OnMissing.PASS) == ref

    with pytest.raises(ValueError):
        ref.resolve(memory_jobstore, on_missing=OnMissing.ERROR)

    # resolve using store
    memory_jobstore.update({"uuid": "123", "index": 1, "output": 101})
    assert ref.resolve(store=memory_jobstore) == 101

    # resolve using store and empty cache
    cache = {}
    assert ref.resolve(store=memory_jobstore, cache=cache) == 101
    assert cache["123"][1] == 101

    # check cache supersedes store
    cache = {"123": {1: "xyz"}}
    assert ref.resolve(store=memory_jobstore, cache=cache) == "xyz"
    assert cache["123"][1] == "xyz"

    # test indexing
    ref = OutputReference("123", (("i", "a"), ("i", 1)))
    memory_jobstore.update({"uuid": "123", "index": 1, "output": {"a": [5, 6, 7]}})
    assert ref.resolve(memory_jobstore) == 6

    # test attribute access
    ref = OutputReference("123", (("a", "__module__"),))
    memory_jobstore.update({"uuid": "123", "index": 1, "output": OutputReference})
    assert ref.resolve(memory_jobstore) == "jobflow.core.reference"

    # test missing attribute throws error
    ref = OutputReference("123", (("a", "b"),))
    memory_jobstore.update({"uuid": "123", "index": 1, "output": [1234]})
    with pytest.raises(AttributeError):
        ref.resolve(memory_jobstore)

    # test missing index throws error
    ref = OutputReference("123", (("i", "b"),))
    with pytest.raises(TypeError):
        ref.resolve(memory_jobstore)


def test_resolve_references(memory_jobstore):
    from jobflow import OnMissing, OutputReference
    from jobflow.core.reference import resolve_references

    # resolve single using store
    ref = OutputReference("123")
    memory_jobstore.update({"uuid": "123", "index": 1, "output": "xyz"})
    output = resolve_references([ref], memory_jobstore)
    assert len(output) == 1
    assert output[ref] == "xyz"

    # resolve multiple using cache
    ref1 = OutputReference("123")
    ref2 = OutputReference("1234")
    memory_jobstore.update({"uuid": "1234", "index": 1, "output": 101})
    output = resolve_references([ref1, ref2], memory_jobstore)
    assert len(output) == 2
    assert output[ref1] == "xyz"
    assert output[ref2] == 101

    # resolve group using cache
    ref1 = OutputReference("123", (("i", "a"),))
    ref2 = OutputReference("123", (("i", "b"),))
    ref3 = OutputReference("1234")
    memory_jobstore.update(
        {"uuid": "123", "index": 1, "output": {"a": "xyz", "b": "abc"}}
    )
    output = resolve_references([ref1, ref2, ref3], memory_jobstore)
    assert len(output) == 3
    assert output[ref1] == "xyz"
    assert output[ref2] == "abc"
    assert output[ref3] == 101

    # test on missing
    ref1 = OutputReference("123")
    ref2 = OutputReference("12345")
    memory_jobstore.update({"uuid": "123", "index": 1, "output": "xyz"})
    output = resolve_references(
        [ref1, ref2], memory_jobstore, on_missing=OnMissing.PASS
    )
    assert len(output) == 2
    assert output[ref1] == "xyz"
    assert output[ref2] == ref2

    ref2 = OutputReference("12345")
    with pytest.raises(ValueError):
        resolve_references([ref1, ref2], memory_jobstore, on_missing=OnMissing.ERROR)

    # resolve using store and empty cache
    cache = {}
    output = resolve_references([ref], memory_jobstore, cache=cache)
    assert len(output) == 1
    assert output[ref] == "xyz"

    # check cache supersedes store
    cache = {"123": {1: 101}}
    output = resolve_references([ref], memory_jobstore, cache=cache)
    assert len(output) == 1
    assert output[ref] == 101

    # test attributes
    ref = OutputReference("123", (("i", "a"), ("i", 1)))
    memory_jobstore.update({"uuid": "123", "index": 1, "output": {"a": [5, 6, 7]}})
    output = resolve_references([ref], memory_jobstore)
    assert output[ref] == 6

    ref = OutputReference("123", (("a", "__module__"),))
    memory_jobstore.update({"uuid": "123", "index": 1, "output": OutputReference})
    output = resolve_references([ref], memory_jobstore)
    assert output[ref] == "jobflow.core.reference"


def test_find_and_get_references():
    from jobflow.core.reference import OutputReference, find_and_get_references

    ref1 = OutputReference("123")
    ref2 = OutputReference("1234", (("a", "a"),))

    # test single reference
    assert find_and_get_references(ref1) == (ref1,)

    # test list and tuple of references
    assert find_and_get_references([ref1]) == (ref1,)
    assert set(find_and_get_references([ref1, ref2])) == {ref1, ref2}
    assert set(find_and_get_references((ref1, ref2))) == {ref1, ref2}

    # test dictionary dictionary values
    assert find_and_get_references({"a": ref1}) == (ref1,)
    assert set(find_and_get_references({"a": ref1, "b": ref2})) == {ref1, ref2}

    # test nested
    assert set(find_and_get_references({"a": [ref1, ref2]})) == {ref1, ref2}
    assert set(find_and_get_references([{"a": ref1}, {"b": ref2}])) == {ref1, ref2}


def test_find_and_resolve_references(memory_jobstore):
    from jobflow.core.reference import (
        OnMissing,
        OutputReference,
        find_and_resolve_references,
    )

    ref1 = OutputReference("123")
    ref2 = OutputReference("1234", (("i", "a"),))
    memory_jobstore.update({"uuid": "123", "index": 1, "output": 101})
    memory_jobstore.update({"uuid": "1234", "index": 1, "output": {"a": "xyz", "b": 5}})

    # test no reference
    assert find_and_resolve_references(True, memory_jobstore) is True
    assert find_and_resolve_references("xyz", memory_jobstore) == "xyz"
    assert find_and_resolve_references([101], memory_jobstore) == [101]

    # test single reference
    assert find_and_resolve_references(ref1, memory_jobstore) == 101

    # test list and tuple of references
    assert find_and_resolve_references([ref1], memory_jobstore) == [101]
    assert find_and_resolve_references([ref1, ref2], memory_jobstore) == [101, "xyz"]

    # test dictionary dictionary values
    output = find_and_resolve_references({"a": ref1}, memory_jobstore)
    assert output == {"a": 101}
    output = find_and_resolve_references({"a": ref1, "b": ref2}, memory_jobstore)
    assert output == {
        "a": 101,
        "b": "xyz",
    }

    # test nested
    output = find_and_resolve_references({"a": [ref1, ref2]}, memory_jobstore)
    assert output == {"a": [101, "xyz"]}
    output = find_and_resolve_references([{"a": ref1}, {"b": ref2}], memory_jobstore)
    assert output == [
        {"a": 101},
        {"b": "xyz"},
    ]

    # test store, blank cache
    cache = {}
    output = find_and_resolve_references(
        {"a": [ref1, ref2]}, store=memory_jobstore, cache=cache
    )
    assert output == {"a": [101, "xyz"]}
    assert cache["123"][1] == 101

    # test cache overrides store
    output = find_and_resolve_references(
        {"a": [ref1, ref2]}, store=memory_jobstore, cache={"123": {1: 1}}
    )
    assert output == {"a": [1, "xyz"]}

    # test on missing
    ref3 = OutputReference("12345", (("i", "a"),))
    output = find_and_resolve_references(
        [ref1, ref3], memory_jobstore, on_missing=OnMissing.PASS
    )
    assert output == [101, ref3]
    output = find_and_resolve_references(
        [ref1, ref3], memory_jobstore, on_missing=OnMissing.NONE
    )
    assert output == [101, None]

    with pytest.raises(ValueError):
        find_and_resolve_references(
            [ref1, ref3], memory_jobstore, on_missing=OnMissing.ERROR
        )


def test_circular_resolve(memory_jobstore):
    from jobflow.core.reference import OutputReference

    # test catching circular resolve failure
    ref1 = OutputReference("12345")
    task_data = {"uuid": ref1.uuid, "index": 1, "output": ref1}
    memory_jobstore.update(task_data)
    with pytest.raises(RuntimeError):
        ref1.resolve(memory_jobstore)


def test_reference_in_output(memory_jobstore):
    from jobflow.core.reference import OnMissing, OutputReference

    # test resolvable reference in job output
    ref1 = OutputReference("12345")
    ref2 = OutputReference("56789")
    task_data1 = {"uuid": ref1.uuid, "index": 1, "output": ref2}
    task_data2 = {"uuid": ref2.uuid, "index": 1, "output": "xyz"}
    memory_jobstore.update(task_data1)
    memory_jobstore.update(task_data2)
    assert "xyz" == ref1.resolve(memory_jobstore)

    # test missing reference in output
    ref1 = OutputReference("12345")
    ref2 = OutputReference("999")
    task_data = {"uuid": ref1.uuid, "index": 1, "output": ref2}
    memory_jobstore.update(task_data)
    assert ref1.resolve(memory_jobstore, on_missing=OnMissing.NONE) is None
    assert ref1.resolve(memory_jobstore, on_missing=OnMissing.PASS) == ref2
    with pytest.raises(ValueError):
        ref1.resolve(memory_jobstore, on_missing=OnMissing.ERROR)



import pytest


def test_apply_mod():
    from jobflow.utils.dict_mods import apply_mod

    d = {"Hello": "World"}
    mod = {"_set": {"Hello": "Universe", "Bye": "World"}}
    apply_mod(mod, d)
    assert d == {"Bye": "World", "Hello": "Universe"}

    mod = {"_unset": {"Hello": 1}}
    apply_mod(mod, d)
    assert d == {"Bye": "World"}

    mod = {"_push": {"List": 1}}
    apply_mod(mod, d)
    assert d == {"Bye": "World", "List": [1]}

    mod = {"_push": {"List": 2}}
    apply_mod(mod, d)
    assert d == {"Bye": "World", "List": [1, 2]}

    mod = {"_inc": {"num": 5}}
    apply_mod(mod, d)
    assert d == {"Bye": "World", "List": [1, 2], "num": 5}

    mod = {"_inc": {"num": 5}}
    apply_mod(mod, d)
    assert d == {"Bye": "World", "List": [1, 2], "num": 10}

    mod = {"_rename": {"num": "number"}}
    apply_mod(mod, d)
    assert d == {"Bye": "World", "List": [1, 2], "number": 10}

    mod = {"_rename": {"num": "number2"}}
    apply_mod(mod, d)
    assert d == {"Bye": "World", "List": [1, 2], "number": 10}

    mod = {"_add_to_set": {"List": 2}}
    apply_mod(mod, d)
    assert d == {"Bye": "World", "List": [1, 2], "number": 10}

    mod = {"_add_to_set": {"List": 3}}
    apply_mod(mod, d)
    assert d == {"Bye": "World", "List": [1, 2, 3], "number": 10}

    e = {}
    mod = {"_add_to_set": {"List": 3}}
    apply_mod(mod, e)
    assert e == {"List": 3}

    mod = {"_add_to_set": {"number": 3}}
    with pytest.raises(ValueError):
        apply_mod(mod, d)

    mod = {"_pull": {"List": 1}}
    apply_mod(mod, d)
    assert d == {"Bye": "World", "List": [2, 3], "number": 10}

    mod = {"_pull": {"num": 1}}
    apply_mod(mod, d)
    assert d == {"Bye": "World", "List": [2, 3], "number": 10}

    mod = {"_pull": {"number": 3}}
    with pytest.raises(ValueError):
        apply_mod(mod, d)

    mod = {"_pull_all": {"List": [2, 3]}}
    apply_mod(mod, d)
    assert d == {"Bye": "World", "List": [], "number": 10}

    mod = {"_pull_all": {"number": 3}}
    with pytest.raises(ValueError):
        apply_mod(mod, d)

    mod = {"_push_all": {"List": list(range(10))}}
    apply_mod(mod, d)
    assert d == {"Bye": "World", "List": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9], "number": 10}

    e = {}
    apply_mod({"_push_all": {"k": list(range(3))}}, e)
    assert e == {"k": [0, 1, 2]}

    mod = {"_pop": {"List": 1}}
    apply_mod(mod, d)
    assert d == {"Bye": "World", "List": [0, 1, 2, 3, 4, 5, 6, 7, 8], "number": 10}

    mod = {"_pop": {"List": -1}}
    apply_mod(mod, d)
    assert d == {"Bye": "World", "List": [1, 2, 3, 4, 5, 6, 7, 8], "number": 10}

    mod = {"_pop": {"List": 2}}
    apply_mod(mod, d)
    assert d == {"Bye": "World", "List": [1, 2, 3, 4, 5, 6, 7, 8], "number": 10}

    mod = {"_pop": {"number": -1}}
    with pytest.raises(ValueError):
        apply_mod(mod, d)

    d = {}
    mod = {"_set": {"a->b->c": 100}}
    apply_mod(mod, d)
    assert d == {"a": {"b": {"c": 100}}}

    mod = {"_set": {"a->b->d": 200}}
    apply_mod(mod, d)
    assert d == {"a": {"b": {"c": 100, "d": 200}}}

    mod = {"_set": {"a->b->d": 300}}
    apply_mod(mod, d)
    assert d == {"a": {"b": {"c": 100, "d": 300}}}

    mod = {"_unset": {"a->b->d": 300}}
    apply_mod(mod, d)
    assert d == {"a": {"b": {"c": 100}}}

    mod = {"_push": {"a->e->f": 300}}
    apply_mod(mod, d)
    assert d == {"a": {"b": {"c": 100}, "e": {"f": [300]}}}

    mod = {"_push_all": {"a->e->f": [100, 200]}}
    apply_mod(mod, d)
    assert d == {"a": {"b": {"c": 100}, "e": {"f": [300, 100, 200]}}}

    mod = {"_inc": {"a->b->c": 2}}
    apply_mod(mod, d)
    assert d == {"a": {"b": {"c": 102}, "e": {"f": [300, 100, 200]}}}

    mod = {"_pull": {"a->e->f": 300}}
    apply_mod(mod, d)
    assert d == {"a": {"b": {"c": 102}, "e": {"f": [100, 200]}}}

    mod = {"_pull_all": {"a->e->f": [100, 200]}}
    apply_mod(mod, d)
    assert d == {"a": {"b": {"c": 102}, "e": {"f": []}}}

    mod = {"_push_all": {"a->e->f": [101, 201, 301, 401]}}
    apply_mod(mod, d)
    assert d == {"a": {"b": {"c": 102}, "e": {"f": [101, 201, 301, 401]}}}

    mod = {"_pop": {"a->e->f": 1}}
    apply_mod(mod, d)
    assert d == {"a": {"b": {"c": 102}, "e": {"f": [101, 201, 301]}}}

    mod = {"_pop": {"a->e->f": -1}}
    apply_mod(mod, d)
    assert d == {"a": {"b": {"c": 102}, "e": {"f": [201, 301]}}}

    mod = {"_abcd": {"a": "b"}}
    with pytest.raises(ValueError):
        apply_mod(mod, d)

    mod = {"_set": {"": ""}}
    with pytest.raises(TypeError):
        apply_mod(mod, d)
